# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import base64
import json
from datetime import datetime

from integration.integration_test_utils import _v3io_parse_get_item_response


def test_v3io_parse_get_item_response():
    request = json.dumps(
        {
            "Item": {
                "int": {"N": "55"},
                "float": {"N": "55.4"},
                "string": {"S": "der die das"},
                "boolean": {"BOOL": True},
                "blob": {"B": base64.b64encode(b"message in a bottle").decode("ascii")},
                "timestamp": {"TS": "1594289596:123456"},
            }
        }
    )
    response = _v3io_parse_get_item_response(request)
    expected = {
        "int": 55,
        "float": 55.4,
        "string": "der die das",
        "boolean": True,
        "blob": b"message in a bottle",
        "timestamp": datetime(2020, 7, 9, 10, 13, 16, 124),
    }
    assert response == expected



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import asyncio
import datetime
import os
import random as rand
import sys
import uuid

import aiohttp
import pandas as pd
import pytest
import v3io

from integration.integration_test_utils import V3ioHeaders, _generate_table_name
from storey import (
    AsyncEmitSource,
    CSVSource,
    CSVTarget,
    DataframeSource,
    FlatMap,
    Map,
    ParquetSource,
    ParquetTarget,
    Reduce,
    ReduceToDataFrame,
    SyncEmitSource,
    build_flow,
)
from storey.dtypes import V3ioError
from storey.utils import get_remaining_path


@pytest.fixture()
def v3io_create_csv():
    # Setup
    file_path = _generate_table_name("bigdata/storey_ci/csv_test") + "file.csv"

    asyncio.run(_write_test_csv(file_path))

    # Test runs
    yield file_path

    # Teardown
    asyncio.run(_delete_file(file_path))


@pytest.fixture()
def v3io_teardown_file():
    # Setup
    file_path = _generate_table_name("bigdata/storey_ci/csv_test")

    # Test runs
    yield file_path

    # Teardown
    asyncio.run(_delete_file(file_path))


async def _write_test_csv(file_path):
    connector = aiohttp.TCPConnector()
    v3io_access = V3ioHeaders()
    client_session = aiohttp.ClientSession(connector=connector)
    try:
        data = "n,n*10\n0,0\n1,10\n2,20\n3,30\n4,40\n5,50\n6,60\n7,70\n8,80\n9,90\n"
        await client_session.put(
            f"{v3io_access._webapi_url}/{file_path}",
            data=data,
            headers=v3io_access._get_put_file_headers,
            ssl=False,
        )
    finally:
        await client_session.close()


async def _delete_file(path):
    connector = aiohttp.TCPConnector()
    v3io_access = V3ioHeaders()
    client_session = aiohttp.ClientSession(connector=connector)
    try:
        response = await client_session.delete(
            f"{v3io_access._webapi_url}/{path}",
            headers=v3io_access._get_put_file_headers,
            ssl=False,
        )
        if response.status >= 300 and response.status != 404 and response.status != 409:
            body = await response.text()
            raise V3ioError(f"Failed to delete item at {path}. Response status code was {response.status}: {body}")
    finally:
        await client_session.close()


def test_csv_reader_from_v3io(v3io_create_csv):
    controller = build_flow(
        [
            CSVSource(f"v3io:///{v3io_create_csv}", header=True),
            FlatMap(lambda x: x),
            Map(lambda x: int(x)),
            Reduce(0, lambda acc, x: acc + x),
        ]
    ).run()

    termination_result = controller.await_termination()
    assert termination_result == 495


def test_csv_reader_from_v3io_error_on_file_not_found():
    controller = build_flow(
        [
            CSVSource("v3io:///bigdatra/tests/idontexist.csv", header=True),
        ]
    ).run()

    with pytest.raises(FileNotFoundError):
        controller.await_termination()


async def async_test_write_csv_to_v3io(v3io_teardown_csv):
    controller = build_flow(
        [
            AsyncEmitSource(),
            CSVTarget(f"v3io:///{v3io_teardown_csv}", columns=["n", "n*10"], header=True),
        ]
    ).run()

    for i in range(10):
        await controller.emit([i, 10 * i])

    await controller.terminate()
    await controller.await_termination()

    v3io_access = V3ioHeaders()
    v3io_client = v3io.aio.dataplane.Client(endpoint=v3io_access._webapi_url, access_key=v3io_access._access_key)
    try:
        container, path = v3io_teardown_csv.split("/", 1)
        result = await v3io_client.object.get(container, path)
    finally:
        await v3io_client.close()

    expected = "n,n*10\n0,0\n1,10\n2,20\n3,30\n4,40\n5,50\n6,60\n7,70\n8,80\n9,90\n"
    assert result.body.decode("utf-8") == expected


def test_write_csv_to_v3io(v3io_teardown_file):
    asyncio.run(async_test_write_csv_to_v3io(v3io_teardown_file))


def test_write_csv_with_dict_to_v3io(v3io_teardown_file):
    file_path = f"v3io:///{v3io_teardown_file}"
    controller = build_flow([SyncEmitSource(), CSVTarget(file_path, columns=["n", "n*10"], header=True)]).run()

    for i in range(10):
        controller.emit({"n": i, "n*10": 10 * i})

    controller.terminate()
    controller.await_termination()

    v3io_access = V3ioHeaders()
    v3io_client = v3io.dataplane.Client(endpoint=v3io_access._webapi_url, access_key=v3io_access._access_key)
    try:
        container, path = v3io_teardown_file.split("/", 1)
        result = v3io_client.object.get(container, path)
    finally:
        v3io_client.close()

    expected = "n,n*10\n0,0\n1,10\n2,20\n3,30\n4,40\n5,50\n6,60\n7,70\n8,80\n9,90\n"
    assert result.body.decode("utf-8") == expected


def test_write_csv_infer_columns_without_header_to_v3io(v3io_teardown_file):
    file_path = f"v3io:///{v3io_teardown_file}"
    controller = build_flow([SyncEmitSource(), CSVTarget(file_path)]).run()

    for i in range(10):
        controller.emit({"n": i, "n*10": 10 * i})

    controller.terminate()
    controller.await_termination()

    v3io_access = V3ioHeaders()
    v3io_client = v3io.dataplane.Client(endpoint=v3io_access._webapi_url, access_key=v3io_access._access_key)
    try:
        container, path = v3io_teardown_file.split("/", 1)
        result = v3io_client.object.get(container, path)
    finally:
        v3io_client.close()

    expected = "0,0\n1,10\n2,20\n3,30\n4,40\n5,50\n6,60\n7,70\n8,80\n9,90\n"
    assert result.body.decode("utf-8") == expected


def test_write_csv_from_lists_with_metadata_and_column_pruning_to_v3io(
    v3io_teardown_file,
):
    file_path = f"v3io:///{v3io_teardown_file}"
    controller = build_flow(
        [
            SyncEmitSource(),
            CSVTarget(file_path, columns=["event_key=$key", "n*10"], header=True),
        ]
    ).run()

    for i in range(10):
        controller.emit({"n": i, "n*10": 10 * i}, key=f"key{i}")

    controller.terminate()
    controller.await_termination()

    v3io_access = V3ioHeaders()
    v3io_client = v3io.dataplane.Client(endpoint=v3io_access._webapi_url, access_key=v3io_access._access_key)
    try:
        container, path = v3io_teardown_file.split("/", 1)
        result = v3io_client.object.get(container, path)
    finally:
        v3io_client.close()

    expected = (
        "event_key,n*10\n"
        "key0,0\n"
        "key1,10\n"
        "key2,20\n"
        "key3,30\n"
        "key4,40\n"
        "key5,50\n"
        "key6,60\n"
        "key7,70\n"
        "key8,80\n"
        "key9,90\n"
    )

    assert result.body.decode("utf-8") == expected


def test_write_to_parquet_to_v3io(setup_teardown_test):
    out_dir = f"v3io:///{setup_teardown_test.table_name}"
    columns = ["my_int", "my_string"]
    controller = build_flow(
        [
            SyncEmitSource(),
            ParquetTarget(out_dir, partition_cols="my_int", columns=columns, max_events=1),
        ]
    ).run()

    expected = []
    for i in range(10):
        controller.emit([i, f"this is {i}"])
        expected.append([i, f"this is {i}"])
    expected_in_pyarrow1 = pd.DataFrame(expected, columns=columns)
    expected_in_pyarrow3 = expected_in_pyarrow1.copy()
    expected_in_pyarrow1["my_int"] = expected_in_pyarrow1["my_int"].astype("int32")
    expected_in_pyarrow3["my_int"] = expected_in_pyarrow3["my_int"].astype("category")

    controller.terminate()
    controller.await_termination()

    read_back_df = pd.read_parquet(out_dir, columns=columns)
    assert read_back_df.equals(expected_in_pyarrow1) or read_back_df.equals(expected_in_pyarrow3)


def test_write_to_parquet_to_v3io_single_file_on_termination(setup_teardown_test):
    out_file = f"v3io:///{setup_teardown_test.table_name}/out.parquet"
    columns = ["my_int", "my_string"]
    controller = build_flow([SyncEmitSource(), ParquetTarget(out_file, columns=columns)]).run()

    expected = []
    for i in range(10):
        controller.emit([i, f"this is {i}"])
        expected.append([i, f"this is {i}"])
    expected = pd.DataFrame(expected, columns=columns)
    controller.terminate()
    controller.await_termination()

    read_back_df = pd.read_parquet(out_file, columns=columns)
    assert read_back_df.equals(expected), f"{read_back_df}\n!=\n{expected}"


# ML-775
def test_write_to_parquet_key_hash_partitioning(setup_teardown_test):
    out_dir = f"v3io:///{setup_teardown_test.table_name}/test_write_to_parquet_default_partitioning{uuid.uuid4().hex}/"
    controller = build_flow(
        [
            SyncEmitSource(key_field=1),
            ParquetTarget(out_dir, columns=["my_int", "my_string"], partition_cols=[("$key", 4)]),
        ]
    ).run()

    expected = []
    expected_buckets = [3, 0, 1, 3, 0, 3, 1, 1, 1, 2]
    for i in range(10):
        controller.emit([i, f"this is {i}"])
        expected.append([i, f"this is {i}", expected_buckets[i]])
    expected = pd.DataFrame(expected, columns=["my_int", "my_string", "hash4_key"])
    controller.terminate()
    controller.await_termination()

    read_back_df = pd.read_parquet(out_dir)
    read_back_df["hash4_key"] = read_back_df["hash4_key"].astype("int64")
    read_back_df.sort_values("my_int", inplace=True)
    read_back_df.reset_index(inplace=True, drop=True)
    assert read_back_df.equals(expected), f"{read_back_df}\n!=\n{expected}"


# ML-701
def test_write_to_parquet_to_v3io_force_string_to_timestamp(setup_teardown_test):
    out_file = f"v3io:///{setup_teardown_test.table_name}/out.parquet"
    columns = ["time"]
    controller = build_flow([SyncEmitSource(), ParquetTarget(out_file, columns=[("time", "datetime")])]).run()

    expected = []
    for _ in range(10):
        t = "2021-03-02T19:45:00"
        controller.emit([t])
        expected.append([datetime.datetime.fromisoformat(t)])
    expected = pd.DataFrame(expected, columns=columns)
    controller.terminate()
    controller.await_termination()

    read_back_df = pd.read_parquet(out_file, columns=columns)
    assert read_back_df.equals(expected)


def test_write_to_parquet_to_v3io_with_indices(setup_teardown_test):
    out_file = f"v3io:///{setup_teardown_test.table_name}/test_write_to_parquet_with_indices{uuid.uuid4().hex}.parquet"
    controller = build_flow(
        [
            SyncEmitSource(),
            ParquetTarget(out_file, index_cols="event_key=$key", columns=["my_int", "my_string"]),
        ]
    ).run()

    expected = []
    for i in range(10):
        controller.emit([i, f"this is {i}"], key=f"key{i}")
        expected.append([f"key{i}", i, f"this is {i}"])
    columns = ["event_key", "my_int", "my_string"]
    expected = pd.DataFrame(expected, columns=columns)
    expected.set_index(["event_key"], inplace=True)
    controller.terminate()
    controller.await_termination()

    read_back_df = pd.read_parquet(out_file, columns=columns)
    assert read_back_df.equals(expected), f"{read_back_df}\n!=\n{expected}"


# ML-602
def test_write_to_parquet_to_v3io_with_nulls(setup_teardown_test):
    out_dir = f"v3io:///{setup_teardown_test.table_name}/test_write_to_parquet_to_v3io_with_nulls{uuid.uuid4().hex}/"
    flow = build_flow(
        [
            SyncEmitSource(),
            ParquetTarget(
                out_dir,
                columns=[
                    ("key=$key", "str"),
                    ("my_int", "int"),
                    ("my_string", "str"),
                    ("my_datetime", "datetime"),
                ],
                partition_cols=[],
                max_events=1,
            ),
        ]
    )

    expected = []
    my_time = datetime.datetime(2021, 1, 1, tzinfo=datetime.timezone(datetime.timedelta(hours=5)))

    controller = flow.run()
    controller.emit({"my_int": 0, "my_string": "hello", "my_datetime": my_time}, key="key1")
    # TODO: Expect correct time zone. Can be done in _Writer, but requires fix for ARROW-10511, which is pyarrow>=3.
    expected.append(
        [
            "key1",
            0,
            "hello",
            my_time.astimezone(datetime.timezone(datetime.timedelta())).replace(tzinfo=None),
        ]
    )
    controller.terminate()
    controller.await_termination()

    controller = flow.run()
    controller.emit({}, key="key2")
    expected.append(["key2", None, None, None])
    controller.terminate()
    controller.await_termination()

    read_back_df = pd.read_parquet(out_dir)
    read_back_df.sort_values("key", inplace=True)
    read_back_df.reset_index(inplace=True, drop=True)
    expected = pd.DataFrame(expected, columns=["key", "my_int", "my_string", "my_datetime"])
    assert read_back_df.compare(expected).empty


def append_and_return(lst, x):
    lst.append(x)
    return lst


def test_filter_before_after_non_partitioned(setup_teardown_test):
    columns = ["my_string", "my_time"]

    df = pd.DataFrame(
        [
            ["good", pd.Timestamp("2018-05-07 13:52:37")],
            ["hello", pd.Timestamp("2019-01-26 14:52:37")],
            ["world", pd.Timestamp("2020-05-11 13:52:37")],
        ],
        columns=columns,
    )
    df.set_index("my_string")

    out_file = f"v3io:///{setup_teardown_test.table_name}/before_after_non_partioned/"
    controller = build_flow(
        [
            DataframeSource(df),
            ParquetTarget(out_file, columns=columns, partition_cols=[]),
        ]
    ).run()
    controller.await_termination()

    before = pd.Timestamp("2019-12-01 00:00:00")
    after = pd.Timestamp("2019-01-01 23:59:59.999999")

    controller = build_flow(
        [
            ParquetSource(out_file, end_filter=before, start_filter=after, filter_column="my_time"),
            Reduce([], append_and_return),
        ]
    ).run()
    read_back_result = controller.await_termination()
    expected = [{"my_string": "hello", "my_time": pd.Timestamp("2019-01-26 14:52:37")}]

    assert read_back_result == expected, f"{read_back_result}\n!=\n{expected}"


def test_filter_before_after_partitioned_random(setup_teardown_test):
    low_limit = pd.Timestamp("2018-01-01")
    high_limit = pd.Timestamp("2020-12-31 23:59:59.999999")

    delta = high_limit - low_limit

    seed_value = rand.randrange(sys.maxsize)
    print("Seed value:", seed_value)
    rand.seed(seed_value)

    random_second = rand.randrange(int(delta.total_seconds()))
    middle_limit = low_limit + datetime.timedelta(seconds=random_second)

    print("middle_limit is " + str(middle_limit))

    number_below_middle_limit = rand.randrange(0, 10)

    def create_rand_data(num_elements, low_limit, high_limit, char):
        import datetime

        delta = high_limit - low_limit

        data = []
        for i in range(0, num_elements):
            element = {}
            element["string"] = char + str(i)
            random_second = rand.randrange(int(delta.total_seconds()))
            element["datetime"] = low_limit + datetime.timedelta(seconds=random_second)
            data.append(element)
        return data

    list1 = create_rand_data(number_below_middle_limit, low_limit, middle_limit, "lower")
    list2 = create_rand_data(10 - number_below_middle_limit, middle_limit, high_limit, "higher")
    combined_list = list1 + list2

    df = pd.DataFrame(combined_list)
    print("data frame is " + str(df))

    all_partition_columns = ["$year", "$month", "$day", "$hour", "$minute", "$second"]
    num_part_columns = rand.randrange(1, 6)
    partition_columns = all_partition_columns[:num_part_columns]
    print("partitioned by " + str(partition_columns))

    out_file = f"v3io:///{setup_teardown_test.table_name}/random/"
    controller = build_flow(
        [
            DataframeSource(df),
            ParquetTarget(
                out_file,
                columns=["string", "datetime"],
                partition_cols=partition_columns,
                time_field="datetime",
            ),
        ]
    ).run()

    controller.await_termination()

    controller = build_flow(
        [
            ParquetSource(
                out_file,
                end_filter=high_limit,
                start_filter=middle_limit,
                filter_column="datetime",
            ),
            Reduce([], append_and_return),
        ]
    ).run()
    read_back_result = controller.await_termination()
    print("expecting " + str(10 - number_below_middle_limit) + " to be above middle limit")
    assert (len(read_back_result)) == 10 - number_below_middle_limit

    controller = build_flow(
        [
            ParquetSource(
                out_file,
                end_filter=middle_limit,
                start_filter=low_limit,
                filter_column="datetime",
            ),
            Reduce([], append_and_return),
        ]
    ).run()
    read_back_result = controller.await_termination()
    print("expecting " + str(number_below_middle_limit) + " to be below middle limit")
    assert (len(read_back_result)) == number_below_middle_limit


def test_filter_before_after_partitioned_inner_other_partition(setup_teardown_test):
    columns = ["my_string", "my_time", "my_city"]

    df = pd.DataFrame(
        [
            ["hello", pd.Timestamp("2020-12-31 14:05:00"), "tel aviv"],
            ["world", pd.Timestamp("2018-12-30 09:00:00"), "haifa"],
            ["sun", pd.Timestamp("2019-12-29 09:00:00"), "tel aviv"],
            ["is", pd.Timestamp("2019-06-30 15:00:45"), "hod hasharon"],
            ["shining", pd.Timestamp("2020-02-28 13:00:56"), "hod hasharon"],
        ],
        columns=columns,
    )
    df.set_index("my_string")

    out_file = f"v3io:///{setup_teardown_test.table_name}/inner_other_partition/"
    controller = build_flow(
        [
            DataframeSource(df),
            ParquetTarget(
                out_file,
                columns=columns,
                partition_cols=["$year", "$month", "$day", "$hour", "my_city"],
                time_field="my_time",
            ),
        ]
    ).run()
    controller.await_termination()

    before = pd.Timestamp("2020-12-31 14:00:00")
    after = pd.Timestamp("2019-07-01 00:00:00")

    controller = build_flow(
        [
            ParquetSource(
                out_file,
                end_filter=before,
                start_filter=after,
                filter_column="my_time",
                columns=columns,
            ),
            Reduce([], append_and_return),
        ]
    ).run()
    read_back_result = controller.await_termination()

    expected = [
        {
            "my_string": "sun",
            "my_time": pd.Timestamp("2019-12-29 09:00:00"),
            "my_city": "tel aviv",
        },
        {
            "my_string": "shining",
            "my_time": pd.Timestamp("2020-02-28 13:00:56"),
            "my_city": "hod hasharon",
        },
    ]

    assert read_back_result == expected, f"{read_back_result}\n!=\n{expected}"


def test_filter_before_after_partitioned_outer_other_partition(setup_teardown_test):
    columns = ["my_string", "my_time", "my_city"]

    df = pd.DataFrame(
        [
            ["shining", pd.Timestamp("2020-12-31 14:00:00"), "ramat gan"],
            ["hello", pd.Timestamp("2020-12-30 08:53:00"), "tel aviv"],
            ["beautiful", pd.Timestamp("2020-12-30 09:00:00"), "haifa"],
            ["sun", pd.Timestamp("2020-12-29 09:00:00"), "tel aviv"],
            ["world", pd.Timestamp("2020-12-30 15:00:45"), "hod hasharon"],
            ["is", pd.Timestamp("2020-12-31 13:00:56"), "hod hasharon"],
        ],
        columns=columns,
    )
    df.set_index("my_string")

    out_file = f"v3io:///{setup_teardown_test.table_name}/outer_other_partition/"
    controller = build_flow(
        [
            DataframeSource(df),
            ParquetTarget(
                out_file,
                columns=columns,
                partition_cols=["my_city", "$year", "$month", "$day", "$hour"],
                time_field="my_time",
            ),
        ]
    ).run()
    controller.await_termination()

    start = pd.Timestamp("2020-12-30 08:53:00")
    end = pd.Timestamp("2020-12-31 14:00:00")

    controller = build_flow(
        [
            ParquetSource(
                out_file,
                start_filter=start,
                end_filter=end,
                filter_column="my_time",
                columns=columns,
            ),
            Reduce([], append_and_return),
        ]
    ).run()
    read_back_result = controller.await_termination()
    expected = [
        {
            "my_string": "beautiful",
            "my_time": pd.Timestamp("2020-12-30 09:00:00"),
            "my_city": "haifa",
        },
        {
            "my_string": "world",
            "my_time": pd.Timestamp("2020-12-30 15:00:45"),
            "my_city": "hod hasharon",
        },
        {
            "my_string": "is",
            "my_time": pd.Timestamp("2020-12-31 13:00:56"),
            "my_city": "hod hasharon",
        },
        {
            "my_string": "shining",
            "my_time": pd.Timestamp("2020-12-31 14:00:00"),
            "my_city": "ramat gan",
        },
    ]

    assert read_back_result == expected, f"{read_back_result}\n!=\n{expected}"


def test_filter_by_time_non_partioned(setup_teardown_test):
    columns = ["my_string", "my_time", "my_city"]

    df = pd.DataFrame(
        [
            ["dina", pd.Timestamp("2019-07-01 00:00:00"), "tel aviv"],
            ["uri", pd.Timestamp("2018-12-30 09:00:00"), "tel aviv"],
            ["katya", pd.Timestamp("2020-12-31 14:00:00"), "hod hasharon"],
        ],
        columns=columns,
    )
    df.set_index("my_string")
    path = "/tmp/test_filter_by_time_non_partioned.parquet"
    df.to_parquet(path)
    start = pd.Timestamp("2019-07-01 00:00:00")
    end = pd.Timestamp("2020-12-31 14:00:00")

    controller = build_flow(
        [
            ParquetSource(path, end_filter=end, start_filter=start, filter_column="my_time"),
            Reduce([], append_and_return),
        ]
    ).run()

    read_back_result = controller.await_termination()

    expected = [
        {
            "my_string": "katya",
            "my_time": pd.Timestamp("2020-12-31 14:00:00"),
            "my_city": "hod hasharon",
        }
    ]

    try:
        assert read_back_result == expected, f"{read_back_result}\n!=\n{expected}"
    finally:
        os.remove(path)


def test_empty_filter_result(setup_teardown_test):
    columns = ["my_string", "my_time", "my_city"]

    df = pd.DataFrame(
        [
            ["dina", pd.Timestamp("2019-07-01 00:00:00"), "tel aviv"],
            ["uri", pd.Timestamp("2018-12-30 09:00:00"), "tel aviv"],
            ["katya", pd.Timestamp("2020-12-31 14:00:00"), "hod hasharon"],
        ],
        columns=columns,
    )
    df.set_index("my_string")
    path = "/tmp/test_filter_by_time_non_partioned.parquet"
    df.to_parquet(path)
    start = pd.Timestamp("2022-07-01 00:00:00")
    end = pd.Timestamp("2022-12-31 14:00:00")

    controller = build_flow(
        [
            ParquetSource(path, end_filter=end, start_filter=start, filter_column="my_time"),
            ReduceToDataFrame(index=["my_string"], insert_key_column_as=["my_string"]),
        ]
    ).run()

    read_back_result = controller.await_termination()

    try:
        pd.testing.assert_frame_equal(read_back_result, pd.DataFrame({}))
    finally:
        os.remove(path)


def test_get_path_utils():
    url = "wasbs://mycontainer@myaccount.blob.core.windows.net/path/to/object.csv"
    schema, path = get_remaining_path(url)
    assert path == "mycontainer/path/to/object.csv"
    assert schema == "wasbs"



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from datetime import datetime, timezone
from enum import Enum
from typing import Callable, List, Optional, Union

import numpy

from .aggregation_utils import get_all_raw_aggregates
from .utils import bucketPerWindow, get_one_unit_of_duration, parse_duration

_termination_obj = object()

known_driver_schemes = ["v3io", "redis", "rediss"]


class Event:
    """The basic unit of data in storey. All steps receive and emit events.

    :param body: the event payload, or data
    :param key: Event key. Used by steps that aggregate events by key, such as AggregateByKey. (Optional). Can be list
    :param processing_time: Event processing time. Defaults to the time the event was created, UTC. (Optional)
    :param id: Event identifier. Usually a unique identifier. (Optional)
    :param headers: Request headers (HTTP only) (Optional)
    :param method: Request method (HTTP only) (Optional)
    :param path: Request path (HTTP only) (Optional)
    :param content_type: Request content type (HTTP only) (Optional)
    :param awaitable_result: Generally not passed directly. (Optional)
    :type awaitable_result: AwaitableResult (Optional)
    """

    _serialize_event_marker = "full_event_wrapper"
    _serialize_fields = ["key", "id"]

    def __init__(
        self,
        body: object,
        key: Optional[Union[str, List[str]]] = None,
        processing_time: Union[None, datetime, int, float] = None,
        id: Optional[str] = None,
        headers: Optional[dict] = None,
        method: Optional[str] = None,
        path: Optional[str] = "/",
        content_type=None,
        awaitable_result=None,
    ):
        self.body = body
        self.key = key
        if processing_time is not None and not isinstance(processing_time, datetime):
            if isinstance(processing_time, str):
                processing_time = datetime.fromisoformat(processing_time)
            elif isinstance(processing_time, (int, float)):
                processing_time = datetime.utcfromtimestamp(processing_time)
            else:
                raise TypeError(
                    f"Event processing_time parameter must be a datetime, string, or int. "
                    f"Got {type(processing_time)} instead."
                )
        self.processing_time = processing_time or datetime.now(timezone.utc)
        self.id = id
        self.headers = headers
        self.method = method
        self.path = path
        self.content_type = content_type
        self._awaitable_result = awaitable_result
        self.error = None

    def __eq__(self, other):
        if not isinstance(other, Event):
            return False

        return (
            self.body == other.body
            and self.id == other.id
            and self.headers == other.headers
            and self.method == other.method
            and self.path == other.path
            and self.content_type == other.content_type
        )  # noqa: E127

    def __str__(self):
        return f"Event(id={self.id}, key={str(self.key)}, time={self.time}, body={self.body})"

    @staticmethod
    def wrap_for_serialization(event, record):
        record = {Event._serialize_event_marker: True, "body": record}
        for field in Event._serialize_fields:
            val = getattr(event, field)
            if val is not None:
                if isinstance(val, datetime):
                    val = datetime.isoformat(val)
                record[field] = val
        return record


class V3ioError(Exception):
    pass


class RedisError(Exception):
    pass


class FlowError(Exception):
    pass


class WindowBase:
    def __init__(self, window, period, window_str):
        self.window_millis = window
        self.period_millis = period
        self.window_str = window_str


class FixedWindow(WindowBase):
    """
    Time window representing fixed time interval. The interval will be divided to 10 periods

    :param window: Time window in the format [0-9]+[smhd]
    """

    def __init__(self, window: str):
        window_millis = parse_duration(window)
        WindowBase.__init__(self, window_millis, window_millis / bucketPerWindow, window)

    def get_total_number_of_buckets(self):
        return bucketPerWindow * 2

    def get_window_start_time(self):
        return self.get_current_window()

    def get_current_window(self):
        return int((datetime.now().timestamp() * 1000) / self.window_millis) * self.window_millis

    def get_current_period(self):
        return int((datetime.now().timestamp() * 1000) / self.period_millis) * self.period_millis


class SlidingWindow(WindowBase):
    """
    Time window representing sliding time interval divided to periods.

    :param window: Time window in the format [0-9]+[smhd]
    :param period: Number of buckets to use for the window [0-9]+[smhd]
    """

    def __init__(self, window: str, period: str):
        window_millis, period_millis = parse_duration(window), parse_duration(period)
        if not window_millis % period_millis == 0:
            raise ValueError("period must be a divider of the window")

        WindowBase.__init__(self, window_millis, period_millis, window)

    def get_total_number_of_buckets(self):
        return int(self.window_millis / self.period_millis)

    def get_window_start_time(self):
        return datetime.now().timestamp() * 1000


def get_window_optimal_size_millis(windows_tuples):
    windows_list = []
    for window_tuple in windows_tuples:
        windows_list.append(window_tuple[0])
    return numpy.lcm.reduce(windows_list)


def get_window_optimal_period_millis(windows_tuples):
    windows_list = []
    for window_tuple in windows_tuples:
        windows_list.append(window_tuple[0])
    return numpy.gcd.reduce(windows_list)


class WindowsBase:
    def __init__(self, period, windows):
        self.max_window_millis = windows[-1][0]
        self.smallest_window_millis = windows[0][0]
        self.period_millis = period
        self.windows = windows  # list of tuples of the form (3600000, '1h')
        self.window_millis = get_window_optimal_size_millis(windows)
        self.total_number_of_buckets = int(self.window_millis / self.period_millis)

    def merge(self, new):
        if self.period_millis != new.period_millis:
            raise ValueError("Cannot use different periods for same aggregation")
        found_new_window = False
        for window in new.windows:
            if window not in self.windows:
                self.windows.append(window)
                found_new_window = True
        if found_new_window:
            if self.max_window_millis < new.max_window_millis:
                self.max_window_millis = new.max_window_millis
            if self.smallest_window_millis > new.smallest_window_millis:
                self.smallest_window_millis = new.smallest_window_millis
            if self.total_number_of_buckets < new.total_number_of_buckets:
                self.total_number_of_buckets = new.total_number_of_buckets
            sorted(set(self.windows), key=lambda tup: tup[0])


def sort_windows_and_convert_to_millis(windows):
    if len(windows) == 0:
        raise ValueError("Windows list can not be empty")

    if isinstance(windows[0], str):
        # Validate windows order
        windows_tuples = [(parse_duration(window), window) for window in windows]
        windows_tuples.sort(key=lambda tup: tup[0])
    else:
        # Internally windows can be passed as tuples
        windows_tuples = windows
    return windows_tuples


class FixedWindows(WindowsBase):
    """
    List of time windows representing fixed time intervals.
    For example: 1h will represent 1h windows starting every round hour.

    :param windows: List of time windows in the format [0-9]+[smhd]
    """

    def __init__(self, windows: List[str]):
        windows_tuples = sort_windows_and_convert_to_millis(windows)
        # The period should be a divisor of the unit of the smallest window,
        # for example if the smallest request window is 2h, the period will be 1h / `bucketPerWindow`
        self.smallest_window_unit_millis = get_one_unit_of_duration(windows_tuples[0][1])
        period = get_window_optimal_period_millis(windows_tuples) / bucketPerWindow
        WindowsBase.__init__(self, period, windows_tuples)

    def round_up_time_to_window(self, timestamp):
        return (
            int(timestamp / self.smallest_window_unit_millis) * self.smallest_window_unit_millis
            + self.smallest_window_unit_millis
        )

    def get_period_by_time(self, timestamp):
        return int(timestamp / self.period_millis) * self.period_millis

    def get_window_start_time_by_time(self, timestamp):
        return int(timestamp / self.window_millis) * self.window_millis

    def merge(self, new):
        if isinstance(new, FixedWindows):
            super(FixedWindows, self).merge(new)
        else:
            self.__init__(new.windows)


class SlidingWindows(WindowsBase):
    """
    List of time windows representing sliding time intervals.
    For example: 1h will represent 1h windows starting from the current time.

    :param windows: List of time windows in the format [0-9]+[smhd]
    :param period: Period in the format [0-9]+[smhd]
    """

    def __init__(self, windows: List[str], period: Optional[str] = None):
        windows_tuples = sort_windows_and_convert_to_millis(windows)

        if period:
            period_millis = parse_duration(period)

            # Verify the given period is a divisor of the windows
            for window in windows_tuples:
                if not window[0] % period_millis == 0:
                    raise ValueError(
                        f"Period must be a divisor of every window, but period {period} does not divide {window}"
                    )
        else:
            # The period should be a divisor of the unit of the smallest window,
            # for example if the smallest request window is 2h, the period will be 1h / `bucketPerWindow`
            smallest_window_unit_millis = get_one_unit_of_duration(windows_tuples[0][1])
            period_millis = smallest_window_unit_millis / bucketPerWindow

        WindowsBase.__init__(self, period_millis, windows_tuples)

    def get_window_start_time_by_time(self, timestamp):
        return int(timestamp / self.period_millis) * self.period_millis


class EmissionType(Enum):
    All = 1
    Incremental = 2


class EmitPolicy:
    def __init__(self, emission_type=EmissionType.All):
        self.emission_type = emission_type


class EmitAfterPeriod(EmitPolicy):
    """
    Emit event for next step after each period ends

    :param delay_in_seconds: Delay event emission by seconds (Optional)
    """

    def __init__(self, delay_in_seconds: Optional[int] = 0, emission_type=EmissionType.All):
        self.delay_in_seconds = delay_in_seconds
        EmitPolicy.__init__(self, emission_type)

    @staticmethod
    def name():
        return "afterPeriod"


class EmitAfterWindow(EmitPolicy):
    """
    Emit event for next step after each window ends

    :param delay_in_seconds: Delay event emission by seconds (Optional)
    """

    def __init__(self, delay_in_seconds: Optional[int] = 0, emission_type=EmissionType.All):
        self.delay_in_seconds = delay_in_seconds
        EmitPolicy.__init__(self, emission_type)

    @staticmethod
    def name():
        return "afterWindow"


class EmitAfterMaxEvent(EmitPolicy):
    """
    Emit the Nth event

    :param max_events: Which number of event to emit
    :param timeout_secs: Emit event after timeout expires even if it didn't reach max_events event (Optional)
    """

    def __init__(
        self,
        max_events: int,
        timeout_secs: Optional[int] = None,
        emission_type=EmissionType.All,
    ):
        self.max_events = max_events
        self.timeout_secs = timeout_secs
        EmitPolicy.__init__(self, emission_type)

    @staticmethod
    def name():
        return "maxEvents"


class EmitAfterDelay(EmitPolicy):
    def __init__(self, delay_in_seconds, emission_type=EmissionType.All):
        self.delay_in_seconds = delay_in_seconds
        EmitPolicy.__init__(self, emission_type)

    @staticmethod
    def name():
        return "afterDelay"


class EmitEveryEvent(EmitPolicy):
    """
    Emit every event
    """

    @staticmethod
    def name():
        return "everyEvent"

    pass


def _dict_to_emit_policy(policy_dict):
    mode = policy_dict.pop("mode")
    if mode == EmitEveryEvent.name():
        policy = EmitEveryEvent()
    elif mode == EmitAfterMaxEvent.name():
        if "maxEvents" not in policy_dict:
            raise ValueError("maxEvents parameter must be specified for maxEvents emit policy")
        policy = EmitAfterMaxEvent(policy_dict.pop("maxEvents"))
    elif mode == EmitAfterDelay.name():
        if "delay" not in policy_dict:
            raise ValueError("delay parameter must be specified for afterDelay emit policy")

        policy = EmitAfterDelay(policy_dict.pop("delay"))
    elif mode == EmitAfterWindow.name():
        policy = EmitAfterWindow(delay_in_seconds=policy_dict.pop("delay", 0))
    elif mode == EmitAfterPeriod.name():
        policy = EmitAfterPeriod(delay_in_seconds=policy_dict.pop("delay", 0))
    else:
        raise TypeError(f"unsupported emit policy type: {mode}")

    if policy_dict:
        raise ValueError(f"got unexpected arguments for emit policy: {policy_dict}")

    return policy


class LateDataHandling(Enum):
    Nothing = 1
    Sort_before_emit = 2


class FieldAggregator:
    """
    Field Aggregator represents an set of aggregation features.

    :param name: Name for the feature.
    :param field: Field in the event body to aggregate.
    :param aggr: List of aggregates to apply.
        Valid values are: [count, sum, sqr, avg, max, min, last, first, sttdev, stdvar]
    :param windows: Time windows to aggregate the data by.
    :param aggr_filter: Filter specifying which events to aggregate. (Optional)
    :param max_value: Maximum value for the aggregation (Optional)
    """

    def __init__(
        self,
        name: str,
        field: Union[str, Callable[[Event], object], None],
        aggr: List[str],
        windows: Union[FixedWindows, SlidingWindows],
        aggr_filter: Optional[Callable[[Event], bool]] = None,
        max_value: Optional[float] = None,
    ):
        if aggr_filter is not None and not callable(aggr_filter):
            raise TypeError(f"aggr_filter expected to be callable, got {type(aggr_filter)}")

        if callable(field):
            self.value_extractor = field
        elif isinstance(field, str):
            self.value_extractor = lambda element: element.get(field)

        self.name = name
        self.aggregations = aggr
        self.windows = windows
        self.aggr_filter = aggr_filter
        self.max_value = max_value

    def get_all_raw_aggregates(self):
        return get_all_raw_aggregates(self.aggregations)

    def should_aggregate(self, element):
        if not self.aggr_filter:
            return True

        return self.aggr_filter(element)


class FixedWindowType(Enum):
    CurrentOpenWindow = 1
    LastClosedWindow = 2



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest

from storey import JoinWithTable, NoSqlTarget, Reduce, SyncEmitSource, Table, build_flow
from storey.redis_driver import RedisDriver

from .integration_test_utils import append_return, get_redis_client


@pytest.fixture()
def redis():
    return get_redis_client()


def test_redis_driver_write(redis):
    try:
        table_name = "test_redis_driver_write"

        driver = RedisDriver(redis)
        controller = build_flow([SyncEmitSource(), NoSqlTarget(Table(table_name, driver))]).run()
        controller.emit({"col1": 0}, "key")
        controller.terminate()
        controller.await_termination()

        table_name = f"{table_name}/"
        hash_key = RedisDriver.make_key("storey:", table_name, "key")
        redis_key = RedisDriver._static_data_key(hash_key)

        data = driver.redis.hgetall(redis_key)
        data_strings = {}
        for key, val in data.items():
            if isinstance(key, bytes):
                data_strings[key.decode("utf-8")] = val.decode("utf-8")
            else:
                data_strings[key] = val

        assert data_strings == {"col1": "0"}
    finally:
        for key in driver.redis.scan_iter(f"*storey:{table_name}*"):
            driver.redis.delete(key)


def test_redis_driver_join(redis):
    try:
        table_name = "test_redis_driver_join"

        driver = RedisDriver(redis)
        table = Table(table_name, driver)
        table_name = f"{table_name}/"

        # Create the data we'll join with in Redis.
        hash_key = RedisDriver.make_key("storey:", table_name, "2")
        redis_key = RedisDriver._static_data_key(hash_key)

        driver.redis.hset(redis_key, mapping={"name": "1234"})
        controller = build_flow(
            [
                SyncEmitSource(),
                JoinWithTable(table, lambda x: x["col2"]),
                Reduce([], lambda acc, x: append_return(acc, x)),
            ]
        ).run()

        controller.emit({"col1": 1, "col2": "2"}, "key")
        controller.emit({"col1": 1, "col2": "2"}, "key")
        controller.terminate()
        termination_result = controller.await_termination()

        expected_result = [{"col1": 1, "col2": "2", "name": 1234}, {"col1": 1, "col2": "2", "name": 1234}]

        assert termination_result == expected_result
    finally:
        for key in driver.redis.scan_iter(f"*storey:{table_name}*"):
            driver.redis.delete(key)



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import base64
import json
import os
import random
import re
import string
from datetime import datetime

import aiohttp
import fakeredis
import redis as r

from storey.drivers import NeedsV3ioAccess
from storey.flow import V3ioError
from storey.redis_driver import RedisDriver

_non_int_char_pattern = re.compile(r"[^-0-9]")
test_base_time = datetime.fromisoformat("2020-07-21T21:40:00+00:00")


class V3ioHeaders(NeedsV3ioAccess):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._get_item_headers = {
            "X-v3io-function": "GetItem",
            "X-v3io-session-key": self._access_key,
        }

        self._get_items_headers = {
            "X-v3io-function": "GetItems",
            "X-v3io-session-key": self._access_key,
        }

        self._put_item_headers = {
            "X-v3io-function": "PutItem",
            "X-v3io-session-key": self._access_key,
        }

        self._update_item_headers = {
            "X-v3io-function": "UpdateItem",
            "X-v3io-session-key": self._access_key,
        }

        self._put_records_headers = {
            "X-v3io-function": "PutRecords",
            "X-v3io-session-key": self._access_key,
        }

        self._create_stream_headers = {
            "X-v3io-function": "CreateStream",
            "X-v3io-session-key": self._access_key,
        }

        self._describe_stream_headers = {
            "X-v3io-function": "DescribeStream",
            "X-v3io-session-key": self._access_key,
        }

        self._seek_headers = {
            "X-v3io-function": "Seek",
            "X-v3io-session-key": self._access_key,
        }

        self._get_records_headers = {
            "X-v3io-function": "GetRecords",
            "X-v3io-session-key": self._access_key,
        }

        self._get_put_file_headers = {"X-v3io-session-key": self._access_key}


def append_return(lst, x):
    lst.append(x)
    return lst


def _generate_table_name(prefix="bigdata/storey_ci/Aggr_test"):
    random_table = "".join([random.choice(string.ascii_letters) for i in range(10)])
    return f"{prefix}/{random_table}/"


def get_redis_client(redis_fake_server=None):
    redis_url = os.environ.get("MLRUN_REDIS_URL")
    if redis_url:
        try:
            res = r.cluster.RedisCluster.from_url(redis_url)
            return res
        except r.cluster.RedisClusterException:
            return r.Redis.from_url(redis_url)
    else:
        return fakeredis.FakeRedis(decode_responses=True, server=redis_fake_server)


def remove_redis_table(table_name):
    redis_client = get_redis_client()
    count = 0
    for key in redis_client.scan_iter(f"*storey-test:{table_name}*"):
        redis_client.delete(key)
        count += 1


drivers_list = ["V3ioDriver", "RedisDriver"]


async def create_stream(stream_path):
    v3io_access = V3ioHeaders()
    connector = aiohttp.TCPConnector()
    client_session = aiohttp.ClientSession(connector=connector)
    request_body = json.dumps({"ShardCount": 2, "RetentionPeriodHours": 1})
    response = await client_session.request(
        "POST",
        f"{v3io_access._webapi_url}/{stream_path}/",
        headers=v3io_access._create_stream_headers,
        data=request_body,
        ssl=False,
    )
    assert response.status == 204, f"Bad response {await response.text()} to request {request_body}"


def create_temp_redis_kv(setup_teardown_test):
    # Create the data we'll join with in Redis.
    table_path = setup_teardown_test.table_name
    redis_fake_server = setup_teardown_test.redis_fake_server
    redis_client = get_redis_client(redis_fake_server=redis_fake_server)

    for i in range(1, 10):
        key = RedisDriver.make_key("storey-test:", table_path, i)
        static_key = RedisDriver._static_data_key(key)
        redis_client.hset(static_key, mapping={"age": f"{10 - i}", "color": f"blue{i}"})


async def create_temp_kv(table_path):
    connector = aiohttp.TCPConnector()
    v3io_access = V3ioHeaders()
    client_session = aiohttp.ClientSession(connector=connector)
    for i in range(1, 10):
        request_body = json.dumps({"Item": {"age": {"N": f"{10 - i}"}, "color": {"S": f"blue{i}"}}})
        response = await client_session.request(
            "PUT",
            f"{v3io_access._webapi_url}/{table_path}/{i}",
            headers=v3io_access._put_item_headers,
            data=request_body,
            ssl=False,
        )
        assert response.status == 200, f"Bad response {await response.text()} to request {request_body}"


def _v3io_parse_get_items_response(response_body):
    response_object = json.loads(response_body)
    i = 0
    for item in response_object["Items"]:
        parsed_item = {}
        for name, type_to_value in item.items():
            for typ, value in type_to_value.items():
                val = _convert_nginx_to_python_type(typ, value)
                parsed_item[name] = val
        response_object["Items"][i] = parsed_item
        i = i + 1
    return response_object


# Deletes the entire table
async def recursive_delete(path, v3io_access):
    connector = aiohttp.TCPConnector()
    client_session = aiohttp.ClientSession(connector=connector)

    try:
        has_more = True
        next_marker = ""
        while has_more:
            get_items_body = {"AttributesToGet": "__name", "Marker": next_marker}
            response = await client_session.put(
                f"{v3io_access._webapi_url}/{path}/",
                headers=v3io_access._get_items_headers,
                data=json.dumps(get_items_body),
                ssl=False,
            )
            body = await response.text()
            if response.status == 200:
                res = _v3io_parse_get_items_response(body)
                for item in res["Items"]:
                    await _delete_item(
                        f'{v3io_access._webapi_url}/{path}/{item["__name"]}',
                        v3io_access,
                        client_session,
                    )

                has_more = "NextMarker" in res
                if has_more:
                    next_marker = res["NextMarker"]
            elif response.status == 404:
                break
            else:
                raise V3ioError(f"Failed to delete table {path}. Response status code was {response.status}: {body}")

        await _delete_item(f"{v3io_access._webapi_url}/{path}/", v3io_access, client_session)
    finally:
        await client_session.close()


async def _delete_item(path, v3io_access, client_session):
    response = await client_session.delete(path, headers=v3io_access._get_put_file_headers, ssl=False)
    if response.status >= 300 and response.status != 404 and response.status != 409:
        body = await response.text()
        raise V3ioError(f"Failed to delete item at {path}. Response status code was {response.status}: {body}")


def _v3io_parse_get_item_response(response_body):
    response_object = json.loads(response_body)["Item"]
    for name, type_to_value in response_object.items():
        val = None
        for typ, value in type_to_value.items():
            val = _convert_nginx_to_python_type(typ, value)
        response_object[name] = val
    return response_object


def _convert_nginx_to_python_type(typ, value):
    if typ == "S" or typ == "BOOL":
        return value
    elif typ == "N":
        if _non_int_char_pattern.search(value):
            return float(value)
        else:
            return int(value)
    elif typ == "B":
        return base64.b64decode(value)
    elif typ == "TS":
        splits = value.split(":", 1)
        secs = int(splits[0])
        nanosecs = int(splits[1])
        return datetime.utcfromtimestamp(secs + nanosecs / 1000000000)
    else:
        raise V3ioError(f"Type {typ} in get item response is not supported")



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from datetime import datetime, timedelta

from storey import Filter, Reduce, SyncEmitSource, build_flow
from storey.dtypes import EmissionType, SlidingWindow
from storey.windowed_store import EmitAfterMaxEvent, Window


def append_return(lst, x):
    lst.append(x)
    return lst


def validate_window(expected, window):
    for elem in window:
        key = elem[0]
        data = elem[1]
        for column in data.features:
            index = 0
            for bucket in data.features[column]:
                if len(bucket.data) > 0:
                    assert bucket.data == expected[key][index]

                index = index + 1


def to_millis(t):
    return t.timestamp() * 1000


def test_windowed_flow():
    controller = build_flow(
        [
            SyncEmitSource(),
            Filter(lambda x: x["col1"] > 3),
            Window(
                SlidingWindow("30m", "5m"),
                "time",
                EmitAfterMaxEvent(max_events=3, emission_type=EmissionType.Incremental),
            ),
            Reduce([], lambda acc, x: append_return(acc, x)),
        ]
    ).run()

    base_time = datetime.now()

    for i in range(10):
        data = {"col1": i, "time": base_time + timedelta(minutes=i)}
        controller.emit(data, f"{i % 2}")

    controller.terminate()
    window_list = controller.await_termination()
    assert len(window_list) == 2

    expected_window_1 = {
        "0": {
            0: [(to_millis(base_time + timedelta(minutes=4)), 4)],
            1: [(to_millis(base_time + timedelta(minutes=6)), 6)],
        },
        "1": {0: [(to_millis(base_time + timedelta(minutes=5)), 5)]},
    }

    expected_window_2 = {
        "0": {1: [(to_millis(base_time + timedelta(minutes=8)), 8)]},
        "1": {
            1: [
                (to_millis(base_time + timedelta(minutes=7)), 7),
                (to_millis(base_time + timedelta(minutes=9)), 9),
            ]
        },
    }

    validate_window(expected_window_1, window_list[0])
    validate_window(expected_window_2, window_list[1])
